(declare (standard-bindings) (extended-bindings) (block) (not safe))

(declare (not core))
(include "points.scm")
(declare (core))

;; we have the laplacian as the stiffness term

(define a (lambda (p) 1.0))

;; and we have c=1 (so f better have u added to it)

(define c (lambda (p) 1.0))

;; no Robin conditions

(define gamma (lambda (p) 0.0))

(define u #f)
(define f #f)
(define g #f)
(define u0 #f)
(define description #f)

(define test-data
  ;; a list of thunks that set u, f, g, and u0 to the right thing.
  (list
   (lambda ()

     ;; I'm not sure that the following is correct

     (set! description "initial problem, u=(fl* t t x (fl- 1.0 x) y (fl- 1.0 y))")

     (set! u (lambda (t p)
	       (let ((x (Point-x p))
		     (y (Point-y p)))
		 (fl* t t x (fl- 1.0 x) y (fl- 1.0 y)))))
     
     (set! f (lambda (t p) 
	       (let ((x (Point-x p))
		     (y (Point-y p))) 
		 (fl+ (fl+ (fl* t (fl* t (fl* (fl* x (fl- 1.0 x)) (fl* y (fl- 1.0 y)))))    
			   (fl* (fl* 2.0 (fl* t (fl* x (fl- 1.0 x)) (fl* (fl* y (fl- 1.0 y)))))))
		      (fl+ (fl* 2.0 (fl* t (fl* t (fl* y (fl- 1.0 y)))))
			   (fl* 2.0 (fl* t (fl* t (fl* x (fl- 1.0 x))))))))))
     
     (set! g (lambda (t p)
	       (let ((x (Point-x p))
		     (y (Point-y p)))
		 (fl+  (fl* t (fl* t (fl* y (fl- y 1.0)))) 
		       (fl* t (fl* t (fl* x (fl- x 1.0))))))))
     
     (set! u0 (lambda (p) 0.0)))
   (lambda ()
     ;; The following looks a bit strange because I did it symbolically with my own package.

     (set! description "simplified problem u=(fl* t (fl+ (fl* x (fl* x (fl* (fl- 1. x) (fl- 1. x)))) (fl* y (fl* y (fl* (fl- 1. y) (fl- 1. y))))))")

     (set! u
	   (lambda (t p)
	     (let ((x (Point-x p))
		   (y (Point-y p)))
	       (fl* t (fl+ (fl* x (fl* x (fl* (fl- 1. x) (fl- 1. x))))
			   (fl* y (fl* y (fl* (fl- 1. y) (fl- 1. y)))))))))
     
     (set! f
	   (lambda (t p)
	     (let ((x (Point-x p))
		   (y (Point-y p)))
	       (fl+ (u t p)
		    (fl- ;; u_t
		     (fl+ (fl* x (fl* x (fl* (fl- 1. x) (fl- 1. x)))) (fl* y (fl* y (fl* (fl- 1. y) (fl- 1. y)))))
		     ;; u_xx
		     (fl* t
			  (fl+ (fl+ (fl* (fl- 1. x) (fl- 1. x))
				    (fl* x (fl+ (fl- x 1.)
						(fl- x 1.))))
			       (fl+ (fl+ (fl* (fl- 1. x)
					      (fl- 1. x))
					 (fl* x 
					      (fl+ (fl- x 1.)
						   (fl- x 1.))))
				    (fl* x (fl+ (fl+ (fl- x 1.) (fl- x 1.)) (fl+ (fl+ (fl- x 1.) (fl- x 1.)) (fl* x 2.)))))))
		     ;;u_yy
		     (fl* t
			  (fl+ (fl+ (fl* (fl- 1. y) (fl- 1. y)) (fl* y (fl+ (fl- y 1.) (fl- y 1.))))
			       (fl+ (fl+ (fl* (fl- 1. y) (fl- 1. y)) (fl* y (fl+ (fl- y 1.) (fl- y 1.))))
				    (fl* y (fl+ (fl+ (fl- y 1.) (fl- y 1.)) (fl+ (fl+ (fl- y 1.) (fl- y 1.)) (fl* y 2.))))))))))))

     (set! g (lambda (t p) 0.))

     (set! u0 (lambda (p) 0.0)))

   (lambda ()
     ;;  now we try maple:

     (define (flsquare x)
       (fl* x x))

     (set! description " u=(fl* t (flsquare (fl* x (fl- 1. x) y (fl- 1. y))))")

     (set! u (lambda (t p)
	       (let ((x (Point-x p))
		     (y (Point-y p)))
		 (fl* t (flsquare (fl* x (fl- 1. x) y (fl- 1. y)))))))

     (set! f
	   (lambda (t p)
	     (let ((x (Point-x p))
		   (y (Point-y p)))
	       (fl+ (u t p)
		    (fl-
		     ;; u_t=x^2*(1-x)^2*y^2*(1-y)^2
		     (flsquare (fl* x (fl- 1. x) y (fl- 1. y)))
		     ;; u_xx = 2*t*(1-x)^2*y^2*(1-y)^2-8*t*x*(1-x)*y^2*(1-y)^2+2*t*x^2*y^2*(1-y)^2
		     (fl+ (fl* 2. t (flsquare (fl* (fl- 1. x) y (fl- 1. y))))
			  (fl- (fl* 8. t x (fl- 1. x) (flsquare (fl* y (- 1. y)))))
			  (fl* 2. t (flsquare (fl* x y (fl- 1. y)))))
		     ;; u_yy=2*t*x^2*(1-x)^2*(1-y)^2-8*t*x^2*(1-x)^2*y*(1-y)+2*t*x^2*(1-x)^2*y^2
		     (fl+ (fl* 2. t (flsquare (fl* (fl- 1. y) x (fl- 1. x))))
			  (fl- (fl* 8. t y (fl- 1. y) (flsquare (fl* x (- 1. x)))))
			  (fl* 2. t (flsquare (fl* y x (fl- 1. x))))))))))

     (set! g (lambda (t p) 0.))

     (set! u0 (lambda (p) 0.0)))

   ;;   we do the simplest thing we can imagine
   (lambda ()
     (set! description "u=(fl* t x)")

     (set! u
	   (lambda (t p)
	     (let ((x (Point-x p))
		   (y (Point-y p)))
	       (fl* t x))))

     (set! f
	   (lambda (t p)
	     (let ((x (Point-x p))
		   (y (Point-y p)))
	       (fl+ (u t p)
		    (fl-
		     ;; u_t
		     x
		     ;; u_xx
		     0.
		     ;; u_yy
		     0.)))))

     (set! g
	   (lambda (t p)
	     (let ((x (Point-x p))
		   (y (Point-y p)))
	       (cond ((fl= x 0.)
		      (fl- t))
		     ((fl= x 1.)
		      t)
		     (else
		      0.)))))

     (set! u0 (lambda (p) 0.)))
   ;; and we get even simpler
   (lambda ()
     (set! description "u=t")

     (set! u
	   (lambda (t p)
	     t))

     (set! f
	   (lambda (t p)
	     (fl+ (u t p) 1.)))

     (set! g
	   (lambda (t p)
	     0.))

     (set! u0
	   (lambda (p) 0.)))))

(define T 2.0)

(define K 5)  ;; number of backward difference steps

(define (test-solution initial-triangulation initial-number-of-timesteps number-of-refinements)
  (let loop ((triangulation initial-triangulation)
	     (N initial-number-of-timesteps)
	     (needed-refinements number-of-refinements))
    (let* ((refined-triangulation
	    (refine triangulation))
	   (delta-t
	    (/ T N))
	   (fes
	    (Triangulation->Linear-finite-element-space triangulation))
	   (fine-fes
	    (Triangulation->Linear-finite-element-space refined-triangulation))
	   (coarse->fine
	    (projector->injector (coarse<-fine fes fine-fes)))
	   (L2-inner-product-operator
	    (->Linear-finite-element-operator fine-fes))
	   (U                               ;; (vector-ref U n) is $U^n$ in the coarse space.
	    (parabolic-equation-solution a c f g gamma u0 fes T N K)))
      (add-mass-terms! L2-inner-product-operator (lambda (p) 1.0))
      (let* ((u-at-time-T    ;; in the fine space
	      (interpolate-procedure->Linear-finite-element-vector (lambda (p) (u T p)) fine-fes))
	     (difference     ;; in the fine space
	      (Vector-subtract (Operator-apply coarse->fine (vector-ref U N))
			       u-at-time-T)))
	(display (flsqrt (Vector-inner-product difference L2-inner-product-operator difference)))
	(newline)
	(if (fx< 0 needed-refinements)
	    (loop refined-triangulation
		  (fx* N 2)
		  (fx- needed-refinements 1)))))))

(for-each (lambda (problem)
	    (problem)
	    (pp description)
	    (test-solution coarse-square-triangulation 4 7))
	  test-data)
